import dataclasses
from email.message import EmailMessage

from factory import Dict
from factory import Factory
from factory import Faker
from factory import LazyFunction
from factory import List
from factory import SubFactory
from faker import Faker as OriginalFaker

from beanhub_inbox.data_types import InboxEmail
from beanhub_inbox.processor import EmailFile

fake = OriginalFaker()


@dataclasses.dataclass(frozen=True)
class EmailAttachment:
    content: bytes
    mime_type: str = "application/octet-stream"
    filename: str | None = None


@dataclasses.dataclass(frozen=True)
class MockEmail:
    headers: dict[str, str]
    subject: str
    from_addresses: list[str]
    recipients: list[str]
    tags: list[str] | None = None
    text: EmailAttachment | None = None
    html: EmailAttachment | None = None
    attachments: list[EmailAttachment] | None = None

    def make_msg(self) -> EmailMessage:
        msg = EmailMessage()
        msg["From"] = ", ".join(self.from_addresses)
        msg["To"] = ", ".join(self.recipients)
        msg["Subject"] = self.subject
        date = self.headers.pop("Date")
        if date is not None:
            msg["Date"] = date.strftime("%a, %d %b %Y %H:%M:%S %z")

        content_parts = []
        if self.text is not None:
            content_parts.append(self.text)
        if self.html is not None:
            content_parts.append(self.html)
        if not content_parts:
            raise ValueError("Need to set at least one of text or html")

        for i, part in enumerate(content_parts):
            if i == 0:
                method = msg.set_content
            else:
                method = msg.add_alternative
            main_type, sub_type = part.mime_type.split("/")
            method(part.content, maintype=main_type, subtype=sub_type)

        if self.attachments is not None:
            for attachment in self.attachments:
                main_type, sub_type = attachment.mime_type.split("/")
                msg.add_attachment(
                    attachment.content,
                    maintype=main_type,
                    subtype=sub_type,
                    filename=attachment.filename,
                )
        return msg


class InboxEmailFactory(Factory):
    id = Faker("uuid4")
    message_id = Faker("slug")
    headers = Dict(
        {
            "mock-header": Faker("slug"),
        }
    )
    subject = Faker("sentence")
    from_addresses = List([Faker("email")])
    recipients = List([Faker("email")])
    tags = List([Faker("slug")])

    class Meta:
        model = InboxEmail


class EmailAttachmentFactory(Factory):
    content = LazyFunction(lambda: fake.paragraph().encode("utf8"))
    mime_type = Faker("mime_type")
    filename = None

    class Meta:
        model = EmailAttachment


class MockEmailFactory(Factory):
    subject = Faker("sentence")
    from_addresses = LazyFunction(lambda: [fake.email()])
    recipients = LazyFunction(lambda: [fake.email()])
    headers = Dict(
        {
            "Date": Faker("past_datetime"),
        }
    )
    text = SubFactory(EmailAttachmentFactory, mime_type="text/plain")
    html = SubFactory(EmailAttachmentFactory, mime_type="text/html")
    attachments = None
    tags = None

    class Meta:
        model = MockEmail


class EmailFileFactory(Factory):
    id = Faker("uuid4")
    subject = Faker("sentence")
    filepath = Faker("file_path", extension="eml")
    from_addresses = LazyFunction(lambda: [fake.email()])
    recipients = LazyFunction(lambda: [fake.email()])
    headers = Dict(
        {
            "Mock-Key": Faker("slug"),
        }
    )
    tags = None

    class Meta:
        model = EmailFile



import pathlib

import pytest
from pytest_factoryboy import register

from .factories import InboxEmailFactory

register(InboxEmailFactory)

TEST_PACKAGE_FOLDER = pathlib.Path(__file__).parent
FIXTURE_FOLDER = TEST_PACKAGE_FOLDER / "fixtures"


@pytest.fixture
def fixtures_folder() -> pathlib.Path:
    return FIXTURE_FOLDER



import pathlib

from jinja2.sandbox import SandboxedEnvironment


def as_posix_path(path: pathlib.Path) -> str:
    return pathlib.Path(path).as_posix()


def make_environment():
    env = SandboxedEnvironment()
    env.filters["as_posix_path"] = as_posix_path
    return env



import csv
import dataclasses
import email.message
import email.policy
import json
import logging
import os
import pathlib
import re
import typing
import uuid

import ollama
from jinja2.sandbox import SandboxedEnvironment
from lxml import etree

from .data_types import ArchiveInboxAction
from .data_types import EmailFileMatchRule
from .data_types import ExtractImportAction
from .data_types import IgnoreImportAction
from .data_types import IgnoreInboxAction
from .data_types import ImportConfig
from .data_types import InboxAction
from .data_types import InboxActionType
from .data_types import InboxConfig
from .data_types import InboxDoc
from .data_types import InboxEmail
from .data_types import InboxMatch
from .data_types import InputConfig
from .data_types import OutputColumn
from .data_types import SimpleFileMatch
from .data_types import StrContainsMatch
from .data_types import StrExactMatch
from .data_types import StrMatch
from .data_types import StrOneOfMatch
from .data_types import StrPrefixMatch
from .data_types import StrRegexMatch
from .data_types import StrSuffixMatch
from .llm import build_row_model
from .llm import DEFAULT_COLUMNS
from .llm import extract
from .llm import think
from .templates import make_environment
from .utils import GeneratorResult
from .utils import parse_tags

logger = logging.getLogger(__name__)
BEANHUB_INBOX_DOMAINS = frozenset(
    ["inbox.beanhub.io", "stage-inbox.beanhub.io", "dev-inbox.beanhub.io"]
)
DEFAULT_PROMPT_TEMPLATE = """\
# Instruction

Extract value from the following email content and output to an object with only one field `{{ column.name }}` in JSON.
Think step by step.

# JSON value definition

{{ column.description }}.
{%- if not column.required %}
Output null value if the value is not available.
{%- endif %}
{%- if column.pattern %}
Ensure the value match regular expression `{{ column.pattern }}`
{%- endif %}

# Email content

```
{{ content }}
```
"""


@dataclasses.dataclass(frozen=True)
class RenderedInputConfig:
    input_config: InputConfig


@dataclasses.dataclass(frozen=True)
class EmailFile:
    id: str
    filepath: str
    subject: str
    from_addresses: list[str]
    recipients: list[str]
    headers: dict[str, str]
    tags: list[str]


@dataclasses.dataclass(frozen=True)
class ProcessImportEvent:
    email_file: EmailFile


@dataclasses.dataclass(frozen=True)
class StartProcessingEmail(ProcessImportEvent):
    pass


@dataclasses.dataclass(frozen=True)
class NoMatch(ProcessImportEvent):
    pass


@dataclasses.dataclass(frozen=True)
class MatchImportRule(ProcessImportEvent):
    import_rule_index: int
    import_config: ImportConfig


@dataclasses.dataclass(frozen=True)
class IgnoreEmail(ProcessImportEvent):
    pass


@dataclasses.dataclass(frozen=True)
class CSVRowExists(ProcessImportEvent):
    output_csv: pathlib.Path
    lineno: int


@dataclasses.dataclass(frozen=True)
class StartExtractingColumn(ProcessImportEvent):
    column: OutputColumn


@dataclasses.dataclass(frozen=True)
class StartThinking(ProcessImportEvent):
    column: OutputColumn
    prompt: str


@dataclasses.dataclass(frozen=True)
class UpdateThinking(ProcessImportEvent):
    column: OutputColumn
    piece: str


@dataclasses.dataclass(frozen=True)
class FinishThinking(ProcessImportEvent):
    column: OutputColumn
    thinking: str


@dataclasses.dataclass(frozen=True)
class FinishExtractingColumn(ProcessImportEvent):
    column: OutputColumn
    value: typing.Any


@dataclasses.dataclass(frozen=True)
class FinishExtractingRow(ProcessImportEvent):
    row: dict


def match_str(pattern: StrMatch, value: str | None) -> typing.Tuple[bool, dict | None]:
    if value is None:
        return False, {}
    if isinstance(pattern, str):
        match = re.match(pattern, value)
        if match is None:
            return False, {}
        return True, match.groupdict()
    elif isinstance(pattern, StrExactMatch):
        return value == pattern.equals, {}
    elif isinstance(pattern, StrPrefixMatch):
        return value.startswith(pattern.prefix), {}
    elif isinstance(pattern, StrSuffixMatch):
        return value.endswith(pattern.suffix), {}
    elif isinstance(pattern, StrContainsMatch):
        return pattern.contains in value, {}
    elif isinstance(pattern, StrOneOfMatch):
        if not pattern.regex:
            if not pattern.ignore_case:
                return value in pattern.one_of, {}
            else:
                return value.lower() in frozenset(
                    item.lower() for item in pattern.one_of
                ), {}
        else:
            for item in pattern.one_of:
                match = re.match(
                    item, value, flags=re.IGNORECASE if pattern.ignore_case else 0
                )
                if match is not None:
                    return True, match.groupdict()
            return False, {}
    else:
        raise ValueError(f"Unexpected str match type {type(pattern)}")


def match_inbox_email(inbox_email: InboxEmail, match: InboxMatch) -> bool:
    if match.tags is not None:
        if inbox_email.tags is None:
            return False
        email_tags = frozenset(inbox_email.tags)
        matching_tags = frozenset(match.tags)
        if matching_tags.intersection(email_tags) != matching_tags:
            return False
    if match.subject is not None:
        if re.match(match.subject, inbox_email.subject) is None:
            return False
    if match.headers is not None:
        for key, value in match.headers.items():
            if key not in inbox_email.headers:
                return False
            email_header_value = inbox_email.headers[key]
            if re.match(value, email_header_value) is None:
                return False
    if match.from_address is not None:
        if not any(
            re.match(match.from_address, address, flags=re.IGNORECASE)
            for address in inbox_email.from_addresses
        ):
            return False
    return True


def process_inbox_email(
    template_env: SandboxedEnvironment,
    inbox_email: InboxEmail,
    inbox_configs: list[InboxConfig],
) -> InboxAction | None:
    for config in inbox_configs:
        if config.match is None or match_inbox_email(
            inbox_email=inbox_email, match=config.match
        ):
            if isinstance(config.action, ArchiveInboxAction):
                template_ctx = inbox_email.model_dump(mode="json")
                output_file = template_env.from_string(
                    config.action.output_file
                ).render(**template_ctx)
                return ArchiveInboxAction(
                    type=InboxActionType.archive, output_file=output_file
                )
            elif isinstance(config.action, IgnoreInboxAction):
                return config.action


def walk_dir_files(
    target_dir: pathlib.Path,
) -> typing.Generator[pathlib.Path, None, None]:
    for root, dirs, files in os.walk(target_dir):
        for file in files:
            yield pathlib.Path(root) / file


def render_input_config_match(
    render_str: typing.Callable, match: SimpleFileMatch
) -> SimpleFileMatch:
    if isinstance(match, str):
        return render_str(match)
    elif isinstance(match, StrExactMatch):
        return StrExactMatch(equals=render_str(match.equals))
    elif isinstance(match, StrRegexMatch):
        return StrRegexMatch(regex=render_str(match.regex))
    else:
        raise ValueError(f"Unexpected match type {type(match)}")


def expand_input_loops(
    template_env: SandboxedEnvironment,
    inputs: list[InputConfig],
    omit_token: str,
) -> typing.Generator[RenderedInputConfig, None, None]:
    for input_config in inputs:
        if input_config.loop is not None:
            if not input_config.loop:
                raise ValueError("Loop content cannot be empty")
            loop = input_config.loop
        else:
            loop = [None]

        for values in loop:
            render_str = lambda value: template_env.from_string(value).render(
                **(dict(omit=omit_token) | (values if values is not None else {}))
            )
            rendered_match = render_input_config_match(
                render_str=render_str,
                match=input_config.match,
            )
            yield RenderedInputConfig(
                input_config=InputConfig(
                    match=rendered_match,
                ),
            )


def match_file(
    pattern: SimpleFileMatch, filepath: pathlib.Path | pathlib.PurePath
) -> bool:
    if isinstance(pattern, str):
        return filepath.match(pattern)
    if isinstance(pattern, StrRegexMatch):
        return re.match(pattern.regex, str(filepath)) is not None
    elif isinstance(pattern, StrExactMatch):
        return str(filepath) == pattern.equals
    else:
        raise ValueError(f"Unexpected file match type {type(pattern)}")


def extract_html_text(html: str) -> str:
    parser = etree.HTMLParser()
    tree = etree.fromstring(html, parser)
    # remove unwanted tags such as style
    etree.strip_elements(tree, "style", "script", with_tail=False)
    content = etree.tostring(tree, method="text", encoding="utf8").decode("utf8")
    return "\n".join(
        filter(lambda line: line, (line.strip() for line in content.splitlines()))
    )


def extract_received_for_email(header_value: str) -> str | None:
    match = re.match("from (.+) by (.+) for (.+);", header_value)
    if match is None:
        return None
    return match.group(3)


def split_emails(email_text: str) -> list[str]:
    return list(map(lambda item: item.strip(), email_text.split(",")))


def build_email_file(
    filepath: pathlib.Path,
    parsed_email: email.message.EmailMessage,
) -> EmailFile:
    received = parsed_email.get("Received")
    tags = None
    if received is not None:
        email_address = extract_received_for_email(received)
        # TODO: make it possible for tags to work for email collected outside of BeanHub
        tags = parse_tags(email_address, domains=BEANHUB_INBOX_DOMAINS)
    from_addresses = split_emails(parsed_email["From"])
    recipients = split_emails(parsed_email["To"])
    subject = parsed_email["Subject"]
    return EmailFile(
        id=filepath.stem,
        filepath=str(filepath),
        subject=subject,
        headers=dict(parsed_email),
        from_addresses=from_addresses,
        recipients=recipients,
        tags=tags,
    )


def match_email_file(
    email_file: EmailFile,
    rule: EmailFileMatchRule,
    extra_attrs: dict | None = None,
) -> typing.Tuple[bool, dict]:
    def get_value(key: str):
        nonlocal email_file
        if extra_attrs is not None and key in extra_attrs:
            return extra_attrs[key]
        return getattr(email_file, key, None)

    match_vars = {}
    for key, pattern in rule.model_dump().items():
        if pattern is None:
            continue
        matched, named_group = match_str(getattr(rule, key), get_value(key))
        if not matched:
            return False, {}
        match_vars |= named_group
    return True, match_vars


def extract_json_block(text: str) -> typing.Generator[dict, None, None]:
    for match in re.finditer("```(json\n)?([^`]*)```", text, flags=re.IGNORECASE):
        try:
            yield json.loads(match.group(2))
        except ValueError:
            continue


def perform_extract_action(
    template_env: SandboxedEnvironment,
    email_file: EmailFile,
    parsed_email: email.message.EmailMessage,
    action: ExtractImportAction,
    llm_model: str,
    workdir_path: pathlib.Path,
) -> typing.Generator[ProcessImportEvent, None, None]:
    workdir_path = workdir_path.resolve().absolute()
    output_csv = workdir_path / action.extract.output_csv
    output_csv = output_csv.resolve().absolute()
    if not output_csv.is_relative_to(workdir_path):
        raise ValueError(f"Output CSV file {output_csv} escapes workdir {workdir_path}")
    if output_csv.exists():
        # TODO: extract this
        with output_csv.open("rt") as fo:
            reader = csv.DictReader(fo)
            if "id" not in reader.fieldnames:
                raise ValueError(
                    f"No id column found in the existing output csv file at {output_csv}"
                )
            for index, row in enumerate(reader):
                email_id = row["id"]
                if email_id == email_file.id:
                    logger.info(
                        "Found email %s row %s in output CSV file %s, skip",
                        email_file.id,
                        index + 1,
                        output_csv,
                    )
                    yield CSVRowExists(
                        email_file=email_file,
                        output_csv=output_csv,
                        lineno=index + 1 + 1,
                    )
                    return

    body = parsed_email.get_body()
    if body.get_content_type() == "text/html":
        text = extract_html_text(body.get_content())
    elif body.get_content_type() == "text/text":
        text = body.get_content()
    elif body.get_content_type() == "multipart/related":
        raise ValueError("Email content with embedded image is not supported yet")
    else:
        raise ValueError(
            f"The email {email_file.id} has no no content available for processing"
        )

    # TODO: get template from action or default value
    template = DEFAULT_PROMPT_TEMPLATE
    if action.extract.template is not None:
        template = action.extract.template

    row = {}
    columns = DEFAULT_COLUMNS
    # TODO: we can run all columns at once to speed up if we need to
    for column in columns:
        logger.info(
            'Extracting "%s" (%s type) column value ...',
            column.name,
            column.type.value,
        )
        yield StartExtractingColumn(email_file=email_file, column=column)
        response_model_cls = build_row_model(
            output_columns=[column],
        )

        prompt = template_env.from_string(template).render(
            json_schema=response_model_cls.model_json_schema(),
            content=text,
            column=column,
        )
        logger.debug(
            "Thinking about extracting data for email %s with prompt:\n%s",
            email_file.id,
            prompt,
        )
        messages = [ollama.Message(role="user", content=prompt)]
        yield StartThinking(email_file=email_file, column=column, prompt=prompt)
        think_generator = GeneratorResult(
            think(model=llm_model, messages=messages, stream=True)
        )
        for part in think_generator:
            yield UpdateThinking(
                email_file=email_file, column=column, piece=part.message.content
            )
        yield FinishThinking(
            email_file=email_file, column=column, thinking=think_generator.value.content
        )

        extracted_value = None
        code_block_json_objs = list(extract_json_block(think_generator.value.content))
        for block_json_obj in code_block_json_objs[::-1]:
            if column.name in block_json_obj:
                extracted_value = block_json_obj[column.name]
                logger.info(
                    'Extracted "%s" value %r from thinking output',
                    column.name,
                    extracted_value,
                )
                break

        if extracted_value is None:
            result = extract(
                model=llm_model,
                messages=messages,
                response_model_cls=response_model_cls,
            )

            json_obj = result.model_dump(mode="json")
            extracted_value = json_obj.get(column.name)
            logger.info(
                'Extracted "%s" value %r with structured output',
                column.name,
                extracted_value,
            )

        yield FinishExtractingColumn(
            email_file=email_file,
            column=column,
            value=extracted_value,
        )

        row[column.name] = extracted_value
        if column.name == "valid" and not extracted_value:
            # TODO: find a way to make it possible to define which column is the "valid"
            logger.info(
                "Email %s is not a valid one, skip all other columns",
                email_file.id,
            )
            break

    logger.info(
        "Write email %s row data %s to CSV file %s",
        email_file.id,
        row,
        output_csv,
    )
    yield FinishExtractingRow(
        email_file=email_file,
        row=row,
    )
    if output_csv.exists():
        # TODO: lock file
        with output_csv.open("at+", newline="") as fo:
            writer = csv.DictWriter(
                fo, fieldnames=["id", *(column.name for column in columns)]
            )
            # TODO: sort by id column?
            writer.writerow(dict(id=email_file.id) | row)
    else:
        # TODO: lock file
        output_csv.parent.mkdir(parents=True, exist_ok=True)
        with output_csv.open("wt", newline="") as fo:
            writer = csv.DictWriter(
                fo, fieldnames=["id", *(column.name for column in columns)]
            )
            writer.writeheader()
            writer.writerow(dict(id=email_file.id) | row)


def process_imports(
    inbox_doc: InboxDoc,
    input_dir: pathlib.Path,
    llm_model: str,
    workdir_path: pathlib.Path,
) -> typing.Generator[ProcessImportEvent, None, None]:
    template_env = make_environment()
    omit_token = uuid.uuid4().hex

    expanded_input_configs = list(
        expand_input_loops(
            template_env=template_env, inputs=inbox_doc.inputs, omit_token=omit_token
        ),
    )

    # sort filepaths for deterministic behavior across platforms
    # TODO: this might be a bit slow if the input dir has a tons of files...
    filepaths = sorted(walk_dir_files(input_dir))
    for filepath in filepaths:
        matched_input_config = None
        for input_config_index, rendered_input_config in enumerate(
            expanded_input_configs
        ):
            input_config = rendered_input_config.input_config
            if match_file(input_config.match, filepath):
                matched_input_config = input_config
                logger.info("Matched input config %s", input_config_index)
                break
        if matched_input_config is None:
            # Not interested in this file, skip
            continue

        rel_filepath = filepath.relative_to(input_dir)
        with filepath.open("rb") as fo:
            parsed_email: email.message.EmailMessage = email.message_from_binary_file(
                fo, policy=email.policy.EmailPolicy()
            )
        email_file = build_email_file(filepath=rel_filepath, parsed_email=parsed_email)
        yield StartProcessingEmail(email_file=email_file)

        matched_import_config = None
        matched_import_config_index = None
        for index, import_config in enumerate(inbox_doc.imports):
            if import_config.match is None:
                matched_import_config = import_config
                matched_import_config_index = index
                break
            else:
                email_matched, _ = match_email_file(
                    email_file=email_file,
                    rule=import_config.match,
                )
                if email_matched:
                    matched_import_config = import_config
                    matched_import_config_index = index
                break
        if matched_import_config is None:
            logger.info(
                "No import rule match for email %s at %s, skip",
                email_file.id,
                email_file.filepath,
            )
            yield NoMatch(email_file=email_file)
            continue

        logger.info(
            "Match email %s at %s with import rule %s",
            email_file.id,
            email_file.filepath,
            matched_import_config.name
            if matched_import_config.name is not None
            else matched_import_config_index,
        )
        yield MatchImportRule(
            email_file=email_file,
            import_rule_index=matched_import_config_index,
            import_config=matched_import_config,
        )
        for action in matched_import_config.actions:
            if isinstance(action, ExtractImportAction):
                yield from perform_extract_action(
                    template_env=template_env,
                    email_file=email_file,
                    parsed_email=parsed_email,
                    action=action,
                    llm_model=llm_model,
                    workdir_path=workdir_path,
                )
            elif isinstance(action, IgnoreImportAction):
                logger.info("Ignore email %s", email_file.id)
                yield IgnoreEmail(email_file=email_file)
            else:
                raise ValueError(f"Unexpected action type {type(action)}")



import typing

import pytest

from beanhub_inbox.utils import parse_tags


@pytest.mark.parametrize(
    "email, domains, expected",
    [
        ("fangpen@launchplatform.com", ["inbox.beanhub.io"], None),
        ("fangpen+mybook@inbox.beanhub.io", ["inbox.beanhub.io"], None),
        ("fangpen@inbox.beanhub.io", ["inbox.beanhub.io"], None),
        (
            "fangpen+mybook+tag0+tag1@inbox.beanhub.io",
            ["inbox.beanhub.io"],
            ["tag0", "tag1"],
        ),
        ("fangpen+mybook+tag0+tag1@inbox.beanhub.io", [], None),
        (
            "fangpen+mybook+tag0+tag1@domain1.com",
            ["domain1.com", "domain2.com"],
            ["tag0", "tag1"],
        ),
        ("fangpen+mybook+tag0+tag1@domain3.com", ["domain1.com", "domain2.com"], None),
        (
            "fangpen+mybook+tag0+tag1@InBoX.BeAnHuB.Io",
            ["inbox.beanhub.io"],
            ["tag0", "tag1"],
        ),
        ("fangpen+mybook+tag0@inbox.beanhub.io", ["inbox.beanhub.io"], ["tag0"]),
        (
            "fangpen+mybook+tag0+tag1+tag2@inbox.beanhub.io",
            ["inbox.beanhub.io"],
            ["tag0", "tag1", "tag2"],
        ),
        (
            "fangpen+mybook+tag-0+tag_1@inbox.beanhub.io",
            ["inbox.beanhub.io"],
            ["tag-0", "tag_1"],
        ),
        ("fangpen+mybook++tag1@inbox.beanhub.io", ["inbox.beanhub.io"], ["", "tag1"]),
        (
            "fangpen+mybook+++tag1@inbox.beanhub.io",
            ["inbox.beanhub.io"],
            ["", "", "tag1"],
        ),
    ],
)
def test_parse_tags(
    email: str, domains: typing.Sequence[str], expected: list[str] | None
):
    assert parse_tags(email_address=email, domains=domains) == expected



import datetime
import json
import logging
import typing

import ollama
import pydantic
import pytest
from pytest_mock import MockFixture

from beanhub_inbox.data_types import OutputColumn
from beanhub_inbox.data_types import OutputColumnType
from beanhub_inbox.llm import build_column_field
from beanhub_inbox.llm import build_row_model
from beanhub_inbox.llm import DECIMAL_REGEX
from beanhub_inbox.llm import extract
from beanhub_inbox.llm import LLMResponseBaseModel
from beanhub_inbox.llm import think
from beanhub_inbox.utils import GeneratorResult

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "output_column, expected",
    [
        pytest.param(
            OutputColumn(
                name="desc",
                type=OutputColumnType.str,
                description="summary of the transaction from the invoice or receipt",
            ),
            (
                "desc",
                typing.Annotated[
                    str,
                    pydantic.Field(
                        description="summary of the transaction from the invoice or receipt"
                    ),
                ],
            ),
            id="str",
        ),
        pytest.param(
            OutputColumn(
                name="txn_id",
                type=OutputColumnType.str,
                description="Id of transaction",
                pattern="[0-9]{10}",
            ),
            (
                "txn_id",
                typing.Annotated[
                    str,
                    pydantic.Field(
                        description="Id of transaction", pattern="[0-9]{10}"
                    ),
                ],
            ),
            id="str-with-regex",
        ),
        pytest.param(
            OutputColumn(
                name="amount",
                type=OutputColumnType.decimal,
                description="transaction amount",
            ),
            (
                "amount",
                typing.Annotated[
                    str,
                    pydantic.Field(
                        description="transaction amount", pattern=DECIMAL_REGEX
                    ),
                ],
            ),
            id="decimal",
        ),
        pytest.param(
            OutputColumn(
                name="year",
                type=OutputColumnType.int,
                description="transaction year",
            ),
            (
                "year",
                typing.Annotated[
                    int,
                    pydantic.Field(description="transaction year"),
                ],
            ),
            id="int",
        ),
        pytest.param(
            OutputColumn(
                name="valid",
                type=OutputColumnType.bool,
                description="is this a invoice or something else",
            ),
            (
                "valid",
                typing.Annotated[
                    bool,
                    pydantic.Field(description="is this a invoice or something else"),
                ],
            ),
            id="bool",
        ),
        pytest.param(
            OutputColumn(
                name="date",
                type=OutputColumnType.date,
                description="date of transaction",
            ),
            (
                "date",
                typing.Annotated[
                    datetime.date,
                    pydantic.Field(description="date of transaction"),
                ],
            ),
            id="date",
        ),
        pytest.param(
            OutputColumn(
                name="timestamp",
                type=OutputColumnType.datetime,
                description="timestamp of transaction",
            ),
            (
                "timestamp",
                typing.Annotated[
                    datetime.datetime,
                    pydantic.Field(description="timestamp of transaction"),
                ],
            ),
            id="datetime",
        ),
    ],
)
def test_build_column_field(
    output_column: OutputColumn, expected: tuple[str, typing.Type]
):
    model = pydantic.create_model(
        "TestModel", **dict([build_column_field(output_column)])
    )
    expected_model = pydantic.create_model("TestModel", **dict([expected]))
    assert model.model_json_schema() == expected_model.model_json_schema()


@pytest.mark.parametrize(
    "output_columns, expected",
    [
        (
            [
                OutputColumn(
                    name="desc",
                    type=OutputColumnType.str,
                    description="summary of the transaction from the invoice or receipt",
                ),
                OutputColumn(
                    name="year",
                    type=OutputColumnType.int,
                    description="transaction year",
                ),
            ],
            pydantic.create_model(
                "Row",
                desc=typing.Annotated[
                    str,
                    pydantic.Field(
                        description="summary of the transaction from the invoice or receipt"
                    ),
                ],
                year=typing.Annotated[
                    int, pydantic.Field(description="transaction year")
                ],
            ),
        ),
    ],
)
def test_build_row_model(
    output_columns: list[OutputColumn], expected: typing.Type[LLMResponseBaseModel]
):
    model = build_row_model(output_columns=output_columns)
    assert model.model_json_schema() == expected.model_json_schema()


@pytest.mark.parametrize(
    "model, prompt, end_token",
    [
        ("deepcoder", "What is the result of 1 + 1?", "</think>"),
    ],
)
def test_think(mocker: MockFixture, model: str, prompt: str, end_token: str):
    mock_chat = mocker.patch.object(ollama, "chat")
    mock_chat.return_value = ollama.ChatResponse(
        message=ollama.Message(
            role="assistant",
            content="<think>The result of 1 + 1 is 2</think> The result is 2",
        )
    )

    think_msg = think(
        model=model,
        messages=[ollama.Message(role="user", content=prompt)],
        end_token=end_token,
    )
    assert think_msg.role == "assistant"
    assert think_msg.content.startswith("<think>")
    assert think_msg.content.endswith("</think>")
    logger.info("Think content:\n%s", think_msg.content)
    assert "2" in think_msg.content


@pytest.mark.parametrize(
    "model, prompt, end_token",
    [
        ("deepcoder", "What is the result of 1 + 1?", "</think>"),
    ],
)
def test_think_stream(mocker: MockFixture, model: str, prompt: str, end_token: str):
    def generate_result():
        for chunk in [
            "<think>",
            "The",
            " result",
            " of",
            " 1",
            " +",
            " 1",
            " is",
            " 2",
            "</think>",
            " The",
            " result",
            " is",
            " 2",
        ]:
            yield ollama.ChatResponse(
                message=ollama.Message(role="assistant", content=chunk)
            )

    mock_chat = mocker.patch.object(ollama, "chat")
    mock_chat.return_value = generate_result()

    chunks: list[str] = []
    think_generator = GeneratorResult(
        think(
            model=model,
            messages=[ollama.Message(role="user", content=prompt)],
            end_token=end_token,
            stream=True,
        )
    )
    for part in think_generator:
        assert part.message.role == "assistant"
        chunks.append(part.message.content)
    content = "".join(chunks)
    assert content.startswith("<think>")
    assert content.endswith("</think>")
    logger.info("Think content:\n%s", content)
    assert "2" in content
    assert think_generator.value.content == content
    assert think_generator.value.role == "assistant"


@pytest.mark.parametrize(
    "model, prompt, end_token, expected",
    [
        pytest.param(
            "deepcoder",
            "What is the result of 1 + 1?",
            "</think>",
            2,
            id="deepcoder-simple-math",
        ),
    ],
)
def test_extract(
    mocker: MockFixture, model: str, prompt: str, end_token: str, expected: int
):
    def generate_result():
        for chunk in json.dumps(dict(value=2)):
            yield ollama.ChatResponse(
                message=ollama.Message(role="assistant", content=chunk)
            )

    mock_chat = mocker.patch.object(ollama, "chat")
    mock_chat.side_effect = [
        ollama.ChatResponse(
            message=ollama.Message(
                role="assistant",
                content="<think>The result of 1 + 1 is 2</think> The result is 2",
            )
        ),
        generate_result(),
    ]

    messages = [ollama.Message(role="user", content=prompt)]
    think_message = think(model=model, messages=messages, end_token=end_token)
    messages.append(think_message)

    class CalculationResult(LLMResponseBaseModel):
        value: int

    result = extract(
        model=model, messages=messages, response_model_cls=CalculationResult
    )
    assert result.value == 2



import json
import pathlib
import re
import textwrap

import ollama
import pytest
import yaml
from jinja2.sandbox import SandboxedEnvironment
from pytest_mock import MockerFixture

from .factories import EmailFileFactory
from .factories import InboxEmailFactory
from .factories import MockEmail
from .factories import MockEmailFactory
from beanhub_inbox.data_types import ArchiveInboxAction
from beanhub_inbox.data_types import EmailFileMatchRule
from beanhub_inbox.data_types import ExtractConfig
from beanhub_inbox.data_types import ExtractImportAction
from beanhub_inbox.data_types import IgnoreInboxAction
from beanhub_inbox.data_types import ImportConfig
from beanhub_inbox.data_types import InboxAction
from beanhub_inbox.data_types import InboxActionType
from beanhub_inbox.data_types import InboxConfig
from beanhub_inbox.data_types import InboxDoc
from beanhub_inbox.data_types import InboxEmail
from beanhub_inbox.data_types import InboxMatch
from beanhub_inbox.data_types import InputConfig
from beanhub_inbox.data_types import SimpleFileMatch
from beanhub_inbox.data_types import StrContainsMatch
from beanhub_inbox.data_types import StrExactMatch
from beanhub_inbox.data_types import StrOneOfMatch
from beanhub_inbox.data_types import StrPrefixMatch
from beanhub_inbox.data_types import StrRegexMatch
from beanhub_inbox.data_types import StrSuffixMatch
from beanhub_inbox.processor import EmailFile
from beanhub_inbox.processor import extract_html_text
from beanhub_inbox.processor import extract_json_block
from beanhub_inbox.processor import extract_received_for_email
from beanhub_inbox.processor import match_email_file
from beanhub_inbox.processor import match_file
from beanhub_inbox.processor import match_inbox_email
from beanhub_inbox.processor import match_str
from beanhub_inbox.processor import process_imports
from beanhub_inbox.processor import process_inbox_email
from beanhub_inbox.processor import render_input_config_match


@pytest.fixture
def template_env() -> SandboxedEnvironment:
    return SandboxedEnvironment()


@pytest.mark.parametrize(
    "pattern, value, expected",
    [
        ("^Foo([0-9]+)", "Foo0", (True, {})),
        ("^Foo([0-9]+)", "Foo", (False, {})),
        ("^Foo([0-9]+)", "foo0", (False, {})),
        ("^Foo([0-9]+)", "", (False, {})),
        ("^Foo([0-9]+)", None, (False, {})),
        (
            r"(?P<first_name>\w+) (?P<last_name>\w+)",
            "Malcolm Reynolds",
            (True, {"first_name": "Malcolm", "last_name": "Reynolds"}),
        ),
        (StrPrefixMatch(prefix="Foo"), "Foo", (True, {})),
        (StrPrefixMatch(prefix="Foo"), "Foobar", (True, {})),
        (StrPrefixMatch(prefix="Foo"), "FooBAR", (True, {})),
        (StrPrefixMatch(prefix="Foo"), "xFooBAR", (False, {})),
        (StrPrefixMatch(prefix="Foo"), "", (False, {})),
        (StrPrefixMatch(prefix="Foo"), None, (False, {})),
        (StrSuffixMatch(suffix="Bar"), "Bar", (True, {})),
        (StrSuffixMatch(suffix="Bar"), "fooBar", (True, {})),
        (StrSuffixMatch(suffix="Bar"), "FooBar", (True, {})),
        (StrSuffixMatch(suffix="Bar"), "Foobar", (False, {})),
        (StrSuffixMatch(suffix="Bar"), "FooBarx", (False, {})),
        (StrSuffixMatch(suffix="Bar"), "", (False, {})),
        (StrSuffixMatch(suffix="Bar"), None, (False, {})),
        (StrContainsMatch(contains="Foo"), "Foo", (True, {})),
        (StrContainsMatch(contains="Foo"), "prefix-Foo", (True, {})),
        (StrContainsMatch(contains="Foo"), "Foo-suffix", (True, {})),
        (StrContainsMatch(contains="Foo"), "prefix-Foo-suffix", (True, {})),
        (StrContainsMatch(contains="Foo"), "prefix-Fo-suffix", (False, {})),
        (StrContainsMatch(contains="Foo"), "", (False, {})),
        (StrContainsMatch(contains="Foo"), None, (False, {})),
        (StrOneOfMatch(one_of=["Foo", "Bar"]), "Foo", (True, {})),
        (StrOneOfMatch(one_of=["Foo", "Bar"]), "Bar", (True, {})),
        (StrOneOfMatch(one_of=["Foo", "Bar"]), "Eggs", (False, {})),
        (StrOneOfMatch(one_of=["Foo", "Bar"]), "boo", (False, {})),
        (StrOneOfMatch(one_of=["Foo", "Bar"], ignore_case=True), "bar", (True, {})),
        (
            StrOneOfMatch(one_of=["Foo(.+)", "Bar(.+)"], regex=True),
            "FooBar",
            (True, {}),
        ),
        (StrOneOfMatch(one_of=["Foo(.+)", "Bar(.+)"], regex=True), "Foo", (False, {})),
        (StrOneOfMatch(one_of=["Foo(.+)", "Bar(.+)"], regex=True), "foo", (False, {})),
        (
            StrOneOfMatch(one_of=["Foo(.+)", "Bar(.+)"], regex=True, ignore_case=True),
            "foobar",
            (True, {}),
        ),
        (
            StrOneOfMatch(
                one_of=["Foo(.+)", "Bar(?P<val>.+)", "bar(.+)"],
                regex=True,
                ignore_case=True,
            ),
            "bar1234",
            (True, dict(val="1234")),
        ),
    ],
)
def test_match_str(
    pattern: SimpleFileMatch, value: str | None, expected: tuple[bool, dict]
):
    assert match_str(pattern, value) == expected


@pytest.mark.parametrize(
    "email_file, rule, extra_attrs, expected",
    [
        pytest.param(
            EmailFileFactory(subject="MOCK_SUBJECT"),
            EmailFileMatchRule(subject=StrExactMatch(equals="MOCK_SUBJECT")),
            None,
            (True, {}),
            id="match-subject",
        ),
        pytest.param(
            EmailFileFactory(subject="MOCK_SUBJECT"),
            EmailFileMatchRule(subject="MOCK_(?P<val>.+)"),
            None,
            (True, dict(val="SUBJECT")),
            id="match-subject-regex-capture",
        ),
        pytest.param(
            EmailFileFactory(subject="MOCK_SUBJECT"),
            EmailFileMatchRule(subject=StrExactMatch(equals="OTHER_SUBJECT")),
            None,
            (False, {}),
            id="not-match-subject",
        ),
        pytest.param(
            EmailFileFactory(filepath="/path/to/mock.eml"),
            EmailFileMatchRule(filepath=StrExactMatch(equals="/path/to/mock.eml")),
            None,
            (True, {}),
            id="match-filepath",
        ),
        pytest.param(
            EmailFileFactory(filepath="/path/to/mock.eml"),
            EmailFileMatchRule(filepath=StrExactMatch(equals="/path/to/other.eml")),
            None,
            (False, {}),
            id="not-match-filepath",
        ),
    ],
)
def test_match_email_file(
    email_file: EmailFile,
    rule: EmailFileMatchRule,
    extra_attrs: dict,
    expected: tuple[bool, dict],
):
    assert match_email_file(email_file, rule, extra_attrs=extra_attrs) == expected


@pytest.mark.parametrize(
    "email, match, expected",
    [
        (
            InboxEmailFactory(
                subject="Mock subject",
            ),
            InboxMatch(subject="Mock .*"),
            True,
        ),
        (
            InboxEmailFactory(
                subject="Other subject",
            ),
            InboxMatch(subject="Mock .*"),
            False,
        ),
        (
            InboxEmailFactory(
                tags=["a", "b", "c"],
            ),
            InboxMatch(tags=["a"]),
            True,
        ),
        (
            InboxEmailFactory(
                tags=["a", "b", "c"],
            ),
            InboxMatch(tags=["b"]),
            True,
        ),
        (
            InboxEmailFactory(
                tags=["a", "b", "c"],
            ),
            InboxMatch(tags=["c"]),
            True,
        ),
        (
            InboxEmailFactory(
                tags=["a", "b", "c"],
            ),
            InboxMatch(tags=["a", "b"]),
            True,
        ),
        (
            InboxEmailFactory(
                tags=["a", "b", "c"],
            ),
            InboxMatch(tags=["a", "c"]),
            True,
        ),
        (
            InboxEmailFactory(
                tags=["a", "b", "c"],
            ),
            InboxMatch(tags=["b", "c"]),
            True,
        ),
        (
            InboxEmailFactory(
                tags=["a", "b", "c"],
            ),
            InboxMatch(tags=["a", "b", "c"]),
            True,
        ),
        (
            InboxEmailFactory(
                tags=["a", "b", "c"],
            ),
            InboxMatch(tags=["a", "b", "c", "d"]),
            False,
        ),
        (
            InboxEmailFactory(
                tags=["a", "b", "c"],
            ),
            InboxMatch(tags=["a", "other"]),
            False,
        ),
        (
            InboxEmailFactory(headers=dict(key="value")),
            InboxMatch(headers=dict(key="value")),
            True,
        ),
        (
            InboxEmailFactory(headers=dict(key="value")),
            InboxMatch(headers=dict(key="val.+")),
            True,
        ),
        (
            InboxEmailFactory(headers=dict(key="value")),
            InboxMatch(headers=dict(key="value", eggs="spam")),
            False,
        ),
        (
            InboxEmailFactory(headers=dict(key="value")),
            InboxMatch(headers=dict(key="other")),
            False,
        ),
        (
            InboxEmailFactory(
                from_addresses=["fangpen@launchplatform.com", "hello@fangpenlin.com"]
            ),
            InboxMatch(from_address="fangpen@launchplatform.com"),
            True,
        ),
        (
            InboxEmailFactory(
                from_addresses=["fangpen@launchplatform.com", "hello@fangpenlin.com"]
            ),
            InboxMatch(from_address="hello@fangpenlin.com"),
            True,
        ),
        (
            InboxEmailFactory(
                from_addresses=["fangpen@launchplatform.com", "hello@fangpenlin.com"]
            ),
            InboxMatch(from_address="fangpen@.+"),
            True,
        ),
        (
            InboxEmailFactory(
                from_addresses=["fangpen@launchplatform.com", "hello@fangpenlin.com"]
            ),
            InboxMatch(from_address=".*fangpen.*"),
            True,
        ),
        (
            InboxEmailFactory(
                from_addresses=["fangpen@launchplatform.com", "hello@fangpenlin.com"]
            ),
            InboxMatch(from_address="other"),
            False,
        ),
    ],
)
def test_match_inbox_email(email: InboxEmail, match: InboxMatch, expected: bool):
    assert match_inbox_email(inbox_email=email, match=match) == expected


@pytest.mark.parametrize(
    "email, inbox_configs, expected",
    [
        pytest.param(
            InboxEmailFactory(
                id="mock-id",
                subject="foo",
            ),
            [
                InboxConfig(
                    match=InboxMatch(subject="eggs"),
                    action=ArchiveInboxAction(
                        output_file="path/to/other/{{ id }}.eml",
                    ),
                ),
                InboxConfig(
                    match=InboxMatch(subject="foo"),
                    action=ArchiveInboxAction(
                        output_file="path/to/{{ id }}.eml",
                    ),
                ),
            ],
            ArchiveInboxAction(output_file="path/to/mock-id.eml"),
            id="order",
        ),
        pytest.param(
            InboxEmailFactory(
                id="mock-id",
                subject="foo",
            ),
            [
                InboxConfig(
                    match=InboxMatch(subject="eggs"),
                    action=ArchiveInboxAction(
                        output_file="path/to/other/{{ id }}.eml",
                    ),
                ),
                InboxConfig(
                    action=ArchiveInboxAction(
                        output_file="path/to/{{ id }}.eml",
                    ),
                ),
            ],
            ArchiveInboxAction(output_file="path/to/mock-id.eml"),
            id="match-none",
        ),
        pytest.param(
            InboxEmailFactory(
                id="mock-id",
                message_id="mock-msg-id",
                subject="foo",
                headers=dict(key="value"),
            ),
            [
                InboxConfig(
                    match=InboxMatch(subject="foo"),
                    action=ArchiveInboxAction(
                        output_file="{{ message_id }}/{{ subject }}/{{ headers['key'] }}.eml",
                    ),
                ),
            ],
            ArchiveInboxAction(output_file="mock-msg-id/foo/value.eml"),
            id="render",
        ),
        pytest.param(
            InboxEmailFactory(
                subject="spam",
            ),
            [
                InboxConfig(
                    match=InboxMatch(subject="eggs"),
                    action=ArchiveInboxAction(
                        output_file="path/to/other/{{ id }}.eml",
                    ),
                ),
                InboxConfig(
                    match=InboxMatch(subject="spam"),
                    action=IgnoreInboxAction(type=InboxActionType.ignore),
                ),
                InboxConfig(
                    action=ArchiveInboxAction(
                        output_file="path/to/{{ id }}.eml",
                    ),
                ),
            ],
            IgnoreInboxAction(type=InboxActionType.ignore),
            id="ignore",
        ),
    ],
)
def test_process_inbox_email(
    template_env: SandboxedEnvironment,
    email: InboxEmail,
    inbox_configs: list[InboxConfig],
    expected: InboxAction | None,
):
    assert (
        process_inbox_email(
            template_env=template_env, inbox_email=email, inbox_configs=inbox_configs
        )
        == expected
    )


@pytest.mark.parametrize(
    "pattern, path, expected",
    [
        ("/path/to/*/foo*.csv", "/path/to/bar/foo.csv", True),
        ("/path/to/*/foo*.csv", "/path/to/bar/foo-1234.csv", True),
        ("/path/to/*/foo*.csv", "/path/to/eggs/foo-1234.csv", True),
        ("/path/to/*/foo*.csv", "/path/to/eggs/foo.csv", True),
        ("/path/to/*/foo*.csv", "/path/from/eggs/foo.csv", False),
        ("/path/to/*/foo*.csv", "foo.csv", False),
        (StrRegexMatch(regex=r"^/path/to/([0-9]+)"), "/path/to/0", True),
        (StrRegexMatch(regex=r"^/path/to/([0-9]+)"), "/path/to/0123", True),
        (StrRegexMatch(regex=r"^/path/to/([0-9]+)"), "/path/to/a0123", False),
        (StrExactMatch(equals="foo.csv"), "foo.csv", True),
        (StrExactMatch(equals="foo.csv"), "xfoo.csv", False),
    ],
)
def test_match_file(pattern: SimpleFileMatch, path: str, expected: bool):
    assert match_file(pattern, pathlib.PurePosixPath(path)) == expected


@pytest.mark.parametrize(
    "match, values, expected",
    [
        (
            "inbox-data/connect/{{ foo }}",
            dict(foo="bar.csv"),
            "inbox-data/connect/bar.csv",
        ),
        (
            "inbox-data/connect/eggs.csv",
            dict(foo="bar.csv"),
            "inbox-data/connect/eggs.csv",
        ),
        (
            StrExactMatch(equals="inbox-data/connect/{{ foo }}"),
            dict(foo="bar.csv"),
            StrExactMatch(equals="inbox-data/connect/bar.csv"),
        ),
        (
            StrRegexMatch(regex="inbox-data/connect/{{ foo }}"),
            dict(foo="bar.csv"),
            StrRegexMatch(regex="inbox-data/connect/bar.csv"),
        ),
    ],
)
def test_render_input_config_match(
    template_env: SandboxedEnvironment,
    match: SimpleFileMatch,
    values: dict,
    expected: SimpleFileMatch,
):
    render_str = lambda value: template_env.from_string(value).render(values)
    assert render_input_config_match(render_str=render_str, match=match) == expected


@pytest.mark.parametrize(
    "filename",
    [
        "sample.yaml",
    ],
)
def test_parse_yaml(fixtures_folder: pathlib.Path, filename: str):
    yaml_file = fixtures_folder / filename
    with yaml_file.open("rb") as fo:
        payload = yaml.safe_load(fo)
    doc = InboxDoc.model_validate(payload)
    assert doc


@pytest.mark.parametrize(
    "header_value, expected",
    [
        (
            (
                "from mail-4317.protonmail.ch (mail-4317.protonmail.ch [185.70.43.17])"
                " by inbound-smtp.us-west-2.amazonaws.com with SMTP id n9dtgvp7tq2eoggselt8sr53kd1eglmau0kbn181"
                " for fangpenlin+mybook+tag0+tag1@dev-inbox.beanhub.io;"
                " Sun, 13 Apr 2025 22:48:38 +0000 (UTC)"
            ),
            "fangpenlin+mybook+tag0+tag1@dev-inbox.beanhub.io",
        ),
        (
            "not-relative stuff",
            None,
        ),
        ("", None),
    ],
)
def test_extract_received_for_email(header_value: str, expected: str | None):
    assert extract_received_for_email(header_value) == expected


@pytest.mark.parametrize(
    "text, expected",
    [
        pytest.param(
            textwrap.dedent("""\
            ```json
            {}
            ```
            """),
            [{}],
            id="json-lang-emtpy-dict",
        ),
        pytest.param(
            textwrap.dedent("""\
            ```
            {}
            ```
            """),
            [{}],
            id="no-lang-emtpy-dict",
        ),
        pytest.param(
            textwrap.dedent("""\
            ```json
            {"key": "value"}
            ```
            """),
            [{"key": "value"}],
            id="json-lang-dict",
        ),
        pytest.param(
            textwrap.dedent("""\
            # Section 1
            
            ```json
            {"key0": "value0"}
            ```
            
            # Section 2
            
            ```
            {"key1": "value1"}
            ```
            
            # Section 3
            
            ```{}```
            
            """),
            [{"key0": "value0"}, {"key1": "value1"}, {}],
            id="multiple",
        ),
    ],
)
def test_extract_json_block(text: str, expected: list[dict]):
    assert list(extract_json_block(text)) == expected


@pytest.mark.parametrize(
    "inbox_doc, email_files, think_results, expected",
    [
        pytest.param(
            InboxDoc(
                inputs=[
                    InputConfig(match="*.eml"),
                ],
                imports=[
                    ImportConfig(
                        actions=[
                            ExtractImportAction(
                                extract=ExtractConfig(output_csv="output.csv")
                            )
                        ]
                    )
                ],
            ),
            {
                "mock.eml": MockEmailFactory(),
            },
            dict(
                valid=False,
            ),
            [
                ("StartProcessingEmail", lambda e: e.email_file.id == "mock"),
                ("MatchImportRule", lambda e: e.email_file.id == "mock"),
                ("StartExtractingColumn", lambda e: e.email_file.id == "mock"),
                ("StartThinking", lambda e: e.email_file.id == "mock"),
                ("UpdateThinking", lambda e: e.email_file.id == "mock"),
                ("FinishThinking", lambda e: e.email_file.id == "mock"),
                (
                    "FinishExtractingColumn",
                    lambda e: e.email_file.id == "mock"
                    and e.column.name == "valid"
                    and e.value == False,
                ),
                (
                    "FinishExtractingRow",
                    lambda e: e.email_file.id == "mock" and e.row == dict(valid=False),
                ),
            ],
            id="unconditional-match",
        ),
        pytest.param(
            InboxDoc(
                inputs=[
                    InputConfig(match="*.eml"),
                ],
                imports=[
                    ImportConfig(
                        match=EmailFileMatchRule(
                            subject=StrExactMatch(equals="MOCK_SUBJECT")
                        ),
                        actions=[
                            ExtractImportAction(
                                extract=ExtractConfig(output_csv="output.csv")
                            )
                        ],
                    )
                ],
            ),
            {
                "mock.eml": MockEmailFactory(subject="MOCK_SUBJECT"),
            },
            dict(
                valid=False,
            ),
            [
                ("StartProcessingEmail", lambda e: e.email_file.id == "mock"),
                ("MatchImportRule", lambda e: e.email_file.id == "mock"),
                ("StartExtractingColumn", lambda e: e.email_file.id == "mock"),
                ("StartThinking", lambda e: e.email_file.id == "mock"),
                ("UpdateThinking", lambda e: e.email_file.id == "mock"),
                ("FinishThinking", lambda e: e.email_file.id == "mock"),
                (
                    "FinishExtractingColumn",
                    lambda e: e.email_file.id == "mock"
                    and e.column.name == "valid"
                    and e.value == False,
                ),
                (
                    "FinishExtractingRow",
                    lambda e: e.email_file.id == "mock" and e.row == dict(valid=False),
                ),
            ],
            id="rule-match",
        ),
        pytest.param(
            InboxDoc(
                inputs=[
                    InputConfig(match="*.eml"),
                ],
                imports=[
                    ImportConfig(
                        match=EmailFileMatchRule(subject=StrExactMatch(equals="OTHER")),
                        actions=[
                            ExtractImportAction(
                                extract=ExtractConfig(output_csv="output.csv")
                            )
                        ],
                    )
                ],
            ),
            {
                "mock.eml": MockEmailFactory(subject="MOCK_SUBJECT"),
            },
            dict(
                valid=False,
            ),
            [
                ("StartProcessingEmail", lambda e: e.email_file.id == "mock"),
                ("NoMatch", lambda e: e.email_file.id == "mock"),
            ],
            id="not-match",
        ),
    ],
)
def test_process_imports(
    mocker: MockerFixture,
    tmp_path: pathlib.Path,
    inbox_doc: InboxDoc,
    email_files: dict[str, MockEmail],
    think_results: dict,
    expected: list,
):
    mock_chat = mocker.patch.object(ollama, "chat")

    def chat_side_effect(messages, **kwargs):
        msg = messages[0]
        match = re.search("with only one field `(.+?)`", msg.content)
        key = match.group(1)
        yield ollama.ChatResponse(
            message=ollama.Message(
                role="assistant", content=json.dumps({key: think_results[key]})
            )
        )

    mock_chat.side_effect = chat_side_effect

    for email_path, email_file in email_files.items():
        (tmp_path / email_path).write_text(str(email_file.make_msg()))

    events = list(
        process_imports(
            inbox_doc=inbox_doc,
            input_dir=tmp_path,
            llm_model="deepcoder",
            workdir_path=tmp_path,
        )
    )
    event_types = list(map(lambda event: event.__class__.__name__, events))
    assert event_types == list(map(lambda item: item[0], expected))
    for event, expected_item in zip(events, expected):
        _, validator = expected_item
        if validator is None:
            continue
        assert validator(event)


@pytest.mark.parametrize(
    "html, expected",
    [
        pytest.param(
            textwrap.dedent("""\
            <div style="font-family: Arial, sans-serif; font-size: 14px;"><br></div><div class="protonmail_quote">
                    ------- Forwarded Message -------<br>
                    From: DigitalOcean Support &lt;support@digitalocean.com&gt;<br>
                    Date: On Sunday, September 1st, 2024 at 12:17 AM<br>
                    Subject: [DigitalOcean] Your 2024-08 invoice is available<br>
                    To: Fang-Pen Lin &lt;fangpen@launchplatform.com&gt;<br>
                    <br>
                    <blockquote class="protonmail_quote" type="cite">
                        <div>Usage charges for 2024-08</div>
                    </blockquote>
            </div>
            """),
            textwrap.dedent("""\
            ------- Forwarded Message -------
            From: DigitalOcean Support <support@digitalocean.com>
            Date: On Sunday, September 1st, 2024 at 12:17 AM
            Subject: [DigitalOcean] Your 2024-08 invoice is available
            To: Fang-Pen Lin <fangpen@launchplatform.com>
            Usage charges for 2024-08"""),
            id="basic",
        ),
        pytest.param(
            textwrap.dedent("""\
            first line
            <style>
            h1 { color: red; }
            </style>
            second line
            <div>
                third line
            </div>
            """),
            "first line\nsecond line\nthird line",
            id="style",
        ),
        pytest.param(
            textwrap.dedent("""\
            first line
            <script>
            console.log('hi')
            </script>
            second line
            <div>
                third line
            </div>
            """),
            "first line\nsecond line\nthird line",
            id="script",
        ),
    ],
)
def test_extract_html_text(html: str, expected: str):
    assert extract_html_text(html) == expected



import datetime
import typing

import ollama
import pydantic

from .data_types import OutputColumn
from .data_types import OutputColumnType


DECIMAL_REGEX = "^-?(0|[1-9][0-9]*)(\\.[0-9]+)?$"
LLM_DEFAULT_OPTIONS = dict(temperature=0)
DEFAULT_COLUMNS: list[OutputColumn] = [
    OutputColumn(
        name="valid",
        type=OutputColumnType.bool,
        description="True if this email is for a transaction such as an invoice or receipt, otherwise False",
    ),
    OutputColumn(
        name="desc",
        type=OutputColumnType.str,
        description="The summary of the transaction in a short sentence",
        required=False,
    ),
    OutputColumn(
        name="merchant",
        type=OutputColumnType.str,
        description="Name of the merchant who sent the email if available",
        required=False,
    ),
    OutputColumn(
        name="amount",
        type=OutputColumnType.decimal,
        description="Transaction amount as a decimal string value, do not include dollar sign and please follow the regex format",
        required=False,
    ),
    OutputColumn(
        name="tax",
        type=OutputColumnType.decimal,
        description="Tax amount as a decimal string value, do not include dollar sign and please follow the regex format",
        required=False,
    ),
    OutputColumn(
        name="txn_id",
        type=OutputColumnType.str,
        description="Id of transaction, such as invoice number or receipt number",
        required=False,
    ),
    OutputColumn(
        name="txn_date",
        type=OutputColumnType.date,
        description="The date of transaction if available, in YYYY-MM-DD format",
        required=False,
    ),
]


class LLMResponseBaseModel(pydantic.BaseModel):
    pass


def build_column_field(output_column: OutputColumn) -> (str, typing.Type):
    kwargs = dict(
        description=output_column.description,
        default=... if output_column.required else None,
    )
    annotated_type: typing.Type
    if output_column.type == OutputColumnType.str:
        if output_column.pattern is not None:
            kwargs["pattern"] = output_column.pattern
        value_type = str
    elif output_column.type == OutputColumnType.int:
        value_type = int
    elif output_column.type == OutputColumnType.decimal:
        kwargs["pattern"] = DECIMAL_REGEX
        value_type = str
    elif output_column.type == OutputColumnType.date:
        value_type = datetime.date
    elif output_column.type == OutputColumnType.datetime:
        value_type = datetime.datetime
    elif output_column.type == OutputColumnType.bool:
        value_type = bool
    else:
        raise ValueError(f"Unexpected type {output_column.type}")
    if not output_column.required:
        value_type = typing.Optional[value_type]
    annotated_type = typing.Annotated[value_type, pydantic.Field(**kwargs)]
    return output_column.name, annotated_type


def build_row_model(
    output_columns: list[OutputColumn],
) -> typing.Type[LLMResponseBaseModel]:
    fields = map(build_column_field, output_columns)
    return pydantic.create_model("Row", **dict(fields), __base__=LLMResponseBaseModel)


def _stream_think(
    model: str,
    messages: list[ollama.Message],
    end_token: str | None = None,
    options: dict | None = None,
) -> typing.Generator[ollama.ChatResponse, None, ollama.Message]:
    chunks: list[str] = []
    for part in ollama.chat(
        model=model, messages=messages, options=options, stream=True
    ):
        msg_content = part["message"]["content"]
        yield part
        chunks.append(msg_content)
        if end_token is not None and msg_content == end_token:
            break
    return ollama.Message(role="assistant", content="".join(chunks))


def think(
    model: str,
    messages: list[ollama.Message],
    end_token: str | None = None,
    options: dict | None = None,
    stream: bool = False,
) -> typing.Generator[ollama.ChatResponse, None, ollama.Message] | ollama.Message:
    if options is None:
        options = LLM_DEFAULT_OPTIONS
    if stream:
        return _stream_think(
            model=model, messages=messages, options=options, end_token=end_token
        )
    resp = ollama.chat(model=model, messages=messages, options=options)
    if end_token is not None:
        resp.message.content = resp.message.content.split(end_token, 1)[0] + end_token
    return resp.message


T = typing.TypeVar("T", bound=LLMResponseBaseModel)


def extract(
    model: str,
    messages: list[ollama.Message],
    response_model_cls: typing.Type[T],
    options: dict | None = None,
) -> T:
    if options is None:
        options = LLM_DEFAULT_OPTIONS
    response = ollama.chat(
        model=model,
        messages=messages,
        options=options,
        format=response_model_cls.model_json_schema(),
        stream=True,
    )

    chunks = []
    for part in response:
        msg_content = part.message.content
        chunks.append(msg_content)

    return response_model_cls.model_validate_json("".join(chunks))



import enum
import typing

import pydantic
from pydantic import BaseModel


@enum.unique
class InboxActionType(str, enum.Enum):
    archive = "archive"
    ignore = "ignore"


@enum.unique
class ImportActionType(str, enum.Enum):
    extract = "extract"
    ignore = "ignore"


class InboxBaseModel(BaseModel):
    pass


class InboxMatch(InboxBaseModel):
    tags: list[str] | None = None
    headers: dict[str, str] | None = None
    subject: str | None = None
    from_address: str | None = None


class ArchiveInboxAction(InboxBaseModel):
    output_file: str
    type: typing.Literal[InboxActionType.archive] = pydantic.Field(
        InboxActionType.archive
    )


class IgnoreInboxAction(InboxBaseModel):
    type: typing.Literal[InboxActionType.ignore]


InboxAction = ArchiveInboxAction | IgnoreInboxAction


class InboxConfig(InboxBaseModel):
    action: InboxAction
    match: InboxMatch | None = None


class StrRegexMatch(InboxBaseModel):
    regex: str


class StrExactMatch(InboxBaseModel):
    equals: str


class StrOneOfMatch(InboxBaseModel):
    one_of: list[str]
    regex: bool = False
    ignore_case: bool = False


class StrPrefixMatch(InboxBaseModel):
    prefix: str


class StrSuffixMatch(InboxBaseModel):
    suffix: str


class StrContainsMatch(InboxBaseModel):
    contains: str


SimpleFileMatch = str | StrExactMatch | StrRegexMatch


class InputConfig(InboxBaseModel):
    match: SimpleFileMatch
    loop: list[dict] | None = None


@enum.unique
class OutputColumnType(enum.Enum):
    str = "str"
    int = "int"
    decimal = "decimal"
    bool = "bool"
    date = "date"
    datetime = "datetime"


class OutputColumn(InboxBaseModel):
    name: str
    type: OutputColumnType
    description: str
    pattern: str | None = None
    required: bool = True


class ExtractConfig(InboxBaseModel):
    output_csv: str
    template: str | None = None


class ExtractImportAction(InboxBaseModel):
    type: typing.Literal[ImportActionType.extract] = pydantic.Field(
        ImportActionType.extract
    )
    extract: ExtractConfig


class IgnoreImportAction(InboxBaseModel):
    type: typing.Literal[ImportActionType.ignore]


ImportAction = ExtractImportAction | IgnoreImportAction


StrMatch = (
    str
    | StrPrefixMatch
    | StrSuffixMatch
    | StrExactMatch
    | StrContainsMatch
    | StrOneOfMatch
)


class EmailFileMatchRule(InboxBaseModel):
    filepath: StrMatch | None = None
    subject: StrMatch | None = None


class ImportConfig(InboxBaseModel):
    # Name of import rule, for users to read only
    name: str | None = None
    match: EmailFileMatchRule | None = None
    actions: list[ImportAction]


class InboxDoc(InboxBaseModel):
    inbox: list[InboxConfig] | None = None
    inputs: list[InputConfig] | None = None
    imports: list[ImportConfig] | None = None


class InboxEmail(InboxBaseModel):
    id: str
    message_id: str
    headers: dict[str, str]
    subject: str
    from_addresses: list[str]
    recipients: list[str]
    tags: list[str] | None = None



import typing

from email_validator import validate_email


T = typing.TypeVar("T")
S = typing.TypeVar("S")
R = typing.TypeVar("R")


class GeneratorResult(typing.Generic[T, S, R]):
    def __init__(self, generator: typing.Generator[T, S, R]):
        self.generator: typing.Generator[T, S, R] = generator
        self.value: R | None = None

    def __iter__(self) -> typing.Generator[T, S, R]:
        self.value = yield from self.generator


def parse_tags(email_address: str, domains: typing.Collection[str]) -> list[str] | None:
    email_info = validate_email(email_address, check_deliverability=False)
    domain = email_info.domain.lower()
    if domain not in domains:
        return None
    parts = email_info.local_part.split("+")
    if len(parts) <= 2:
        return None
    return parts[2:]
